/*
								+----------------------------------+
								|                                  |
								|    ***  convex.madness02  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <vector>
#include <algorithm>
#include "../Vector.h"

//const float f_epsilon = 1e-3f; // in Vector.cpp

Vector3f v_mouse(0, 0, 0);
// mouse cursor

int n_width = 640, n_height = 640;
// window size

std::vector<Vector3f> vertex_list;
// polygon and it's vertex list

Vector3f *p_snap_vertex = 0;
// vertex we're holding with mouse

/*
 *	class CSnapVertex
 *		- simple function object for snapping vertices
 */
class CSnapVertex {
protected:
	float m_f_best_distance;
	Vector3f *m_p_snap_vertex;
	Vector3f m_v_mouse;

public:
	inline CSnapVertex(float f_max_distance, Vector3f v_mouse)
		:m_f_best_distance(f_max_distance), m_p_snap_vertex(0), m_v_mouse(v_mouse)
	{}

	inline void operator ()(Vector3f &r_vertex)
	{
		if((m_v_mouse - r_vertex).f_Length() < m_f_best_distance) {
			m_f_best_distance = (m_v_mouse - r_vertex).f_Length();
			m_p_snap_vertex = &r_vertex;
		}
	}

	inline operator Vector3f*() const
	{
		return m_p_snap_vertex;
	}
};

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set ortho projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera

	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	// wireframe

	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	// smooth points and lines

	glPointSize(8);
}

/*
 *	class CDrawPolygon
 *		- simple function object we can use to simply enter polygon's vertices to OpenGL
 */
class CDrawPolygon {
public:
	inline void operator ()(const Vector3f &r_t_vertex) const
	{
		glVertex3f(r_t_vertex.x, r_t_vertex.y, r_t_vertex.z);
	}
};

void RenderString(Vector3f v_pos, const char *p_s_string,
				  void *p_font = GLUT_BITMAP_8_BY_13)
{
	glRasterPos2f(v_pos.x, v_pos.y);
	for(; *p_s_string; ++ p_s_string)
    	glutBitmapCharacter(p_font, *p_s_string);
}

void Draw_Arrow(const char *p_s_name, Vector3f v_org, Vector3f v_tar,
	float f_spike_length = .05f, float f_spike_ratio = .7f)
{
	RenderString(v_org + (v_tar - v_org) * .5f, p_s_name);

	glBegin(GL_LINE_STRIP);
	glVertex2f(v_org.x, v_org.y);
	glVertex2f(v_tar.x, v_tar.y); // telo sipky

	float f_length = (v_tar - v_org).f_Length();
	v_org = v_tar + (v_org - v_tar) * (f_spike_length / f_length); // vychozi bod spice sipky

	Vector3f v_tangent(-(v_tar - v_org).y, (v_tar - v_org).x, 0);
	v_tangent *= f_spike_ratio * .5f; // kolmice na sipku

	glVertex2f(v_org.x + v_tangent.x, v_org.y + v_tangent.y);
	glVertex2f(v_org.x - v_tangent.x, v_org.y - v_tangent.y);
	glVertex2f(v_tar.x, v_tar.y); // spice
	glEnd();
}

void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glColor3f(1, 1, 1);
	// white

	glEnable(GL_BLEND);
	// enable blending

	glBegin(GL_POLYGON);
	std::for_each(vertex_list.begin(), vertex_list.end(), CDrawPolygon());
	glEnd();
	// draw polygon

	glColor3f(1, 0, 0);
	// red
	glBegin(GL_POINTS);
	std::for_each(vertex_list.begin(), vertex_list.end(), CDrawPolygon());
	glEnd();
	// draw polygon's vertices

	glColor3f(1, 1, 0);
	// yellow

	glBegin(GL_POINTS);
	glVertex2f(v_mouse.x, v_mouse.y);
	glEnd();
	// draw mouse

	glColor3f(1, 1, 1);
	// white

	Vector3f v_edge1 = vertex_list[1] - vertex_list[0],
			 v_edge2 = vertex_list[2] - vertex_list[0];
	float f_inv_length2_edge1 = 1.0f / v_edge1.f_Length2();
	Vector3f v_prep = v_edge2 - v_edge1 *
		(v_edge1.f_Dot(v_edge2) * f_inv_length2_edge1);
	float f_rescale = 1 / v_prep.f_Dot(v_edge2);
	Vector3f v_t = v_mouse - vertex_list[0];
	float f_u = v_t.f_Dot(v_prep) * f_rescale;
	float f_v = (v_t - v_edge2 * f_u).f_Dot(v_edge1) * f_inv_length2_edge1;
	// calculate barycentric coordinates for the point

	Draw_Arrow("edge1", vertex_list[0], vertex_list[1]);
	Draw_Arrow("edge2", vertex_list[0], vertex_list[2]);
	Draw_Arrow("perp", vertex_list[0], vertex_list[0] + v_prep);
	Draw_Arrow("edge2.u", vertex_list[0], vertex_list[0] + (v_edge2 * f_u));
	Draw_Arrow("edge1.v", vertex_list[0] + (v_edge2 * f_u),
		vertex_list[0] + (v_edge2 * f_u) + (v_edge1 * f_v));
	// draw arrows inside triangle

	char p_s_str[64];
	sprintf(p_s_str, "u = %.5f", f_u);
	if(f_u >= 0)
		glColor3f(1, 1, 1);
	else
		glColor3f(1, 0, 0);
	RenderString(Vector3f(-1, .95f, 0), p_s_str);
	sprintf(p_s_str, "v = %.5f", f_v);
	if(f_v >= 0)
		glColor3f(1, 1, 1);
	else
		glColor3f(1, 0, 0);
	RenderString(Vector3f(-1, .90f, 0), p_s_str);
	sprintf(p_s_str, "u + v = %.5f", f_u + f_v);
	if(f_u + f_v <= 1)
		glColor3f(1, 1, 1);
	else
		glColor3f(1, 0, 0);
	RenderString(Vector3f(-1, .85f, 0), p_s_str);
	// draw coordinate values

	glDisable(GL_BLEND);

	glutSwapBuffers();
}

void onMouseClick(int button, int state, int x, int y)
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;
	// get current mouse position

	switch(button) {
    case GLUT_LEFT_BUTTON:
		p_snap_vertex = std::for_each(vertex_list.begin(),
			vertex_list.end(), CSnapVertex(.05f, v_mouse));
		// try to snap some vertex near the mouse
		break;
    case GLUT_RIGHT_BUTTON:
    case GLUT_MIDDLE_BUTTON:
		p_snap_vertex = 0; // nothing with those buttons
		break;
	};
}

void onMouseMotion(int x, int y) // dragging
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;

	if(p_snap_vertex)
		*p_snap_vertex = Vector3f(v_mouse.x, v_mouse.y, 0);
	// move our vertex

	glutPostRedisplay();
}

void onPassiveMouseMotion(int x, int y) // just moving mouse
{
	v_mouse.x = (2.0f * x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * y) / n_height;

	glutPostRedisplay();
}

int main(int n_arg_num, /*const*/ char **p_arg_list) // glut doesn't like 'const'
{
	vertex_list.push_back(Vector3f(.7f * sin(.0f), .7f * cos(.0f), 0));
	vertex_list.push_back(Vector3f(.7f * sin(f_pi * 2 / 3), .7f * cos(f_pi * 2 / 3), 0));
	vertex_list.push_back(Vector3f(.7f * sin(f_pi * 4 / 3), .7f * cos(f_pi * 4 / 3), 0));
	// create triangle

	glutInit(&n_arg_num, p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("convex.madness");
	// init OpenGL using GLUT

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	glutPassiveMotionFunc(onPassiveMouseMotion);
	// register callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
