/*
								+----------------------------------+
								|                                  |
								|    ***  Polygon template  ***    |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|            Polygon.h             |
								|                                  |
								+----------------------------------+
*/

/*
 *	02-28-2007
 *
 *	removed error in CPolygon2<TVertStruct>::Cut and CPolygon2<TVertStruct>::Split
 *	in cases edge was longer than 1 treshold for adding new edge vertex was bigger
 *	than epsilon so potencial maximal vertex - plane distance deviation was greater
 *	than epsilon. old code also always generated new vertex even in case it was not
 *	going to be used.
 *	fixed unnecessary / double vertices issue in CPolygon2<TVertStruct>::Split
 *
 */

#ifndef __POLYGON_INCLUDED
#define __POLYGON_INCLUDED

/*
 *	template <class TBaseVertex>
 *	struct TRefVertex
 *		- template for creating reference vertices for any arbitrary vertices
 */
template <class TBaseVertex>
struct TRefVertex {
	TBaseVertex *m_p_ref; // referenced vertex
	std::vector<TBaseVertex*> &m_r_vertex_pool; // vertex pool (here we going to add new vertices)

	typedef TBaseVertex TIntermediate;
	// intermediate type for calculations will be TBaseVertex

	/*
	 *	inline TRefVertex(const TBaseVertex &r_t_vertex, std::vector<TBaseVertex*> &r_vertex_pool)
	 *		- default constructor
	 */
	inline TRefVertex(const TBaseVertex &r_t_vertex, std::vector<TBaseVertex*> &r_vertex_pool)
		:m_p_ref(0), m_r_vertex_pool(r_vertex_pool)
	{
		m_p_ref = new TBaseVertex(r_t_vertex); // alloc a new TBaseVertex
		m_r_vertex_pool.push_back(m_p_ref); // insert it to the list
	}

	/*
	 *	inline TRefVertex(typename std::vector<TBaseVertex*>::iterator p_vertex,
	 *		std::vector<TBaseVertex*> &r_vertex_pool)
	 *		- reference vertex that is already in the list
	 */
	inline TRefVertex(typename std::vector<TBaseVertex*>::iterator p_vertex,
		std::vector<TBaseVertex*> &r_vertex_pool)
		:m_p_ref(*p_vertex), m_r_vertex_pool(r_vertex_pool)
	{}

	/*
	 *	inline TRefVertex(const TRefVertex &r_t_vertex)
	 *		- copy constructor
	 */
	inline TRefVertex(const TRefVertex &r_t_vertex)
		:m_p_ref(r_t_vertex.m_p_ref), m_r_vertex_pool(r_t_vertex.m_r_vertex_pool)
	{}

	/*
	 *	inline TRefVertex operator =(const TRefVertex &r_t_vertex)
	 *		- copy operator = (need it to avoid unwanted conversions)
	 */
	inline TRefVertex operator =(const TRefVertex &r_t_vertex)
	{
		m_p_ref = r_t_vertex.m_p_ref;
		m_r_vertex_pool = r_t_vertex.m_r_vertex_pool;

		return *this;
	}

	/*
	 *	inline TRefVertex operator =(const TBaseVertex &r_t_vertex)
	 *		- conversion from TBaseVertex to TRefVertex (conversion
	 *		  is solved by adding vertex to the vertex list)
	 */
	inline TRefVertex operator =(const TBaseVertex &r_t_vertex)
	{
		m_p_ref = new TBaseVertex(r_t_vertex); // alloc a new TBaseVertex
		m_r_vertex_pool.push_back(m_p_ref); // insert it to the list

		return *this;
	}

	/*
	 *	inline operator TBaseVertex() const
	 *		- now we need means to convert vertex type to intermediate vertex type
	 */
	inline operator TBaseVertex() const
	{
		return *m_p_ref;
	}

	/*
	 *	inline operator Vector3f() const
	 *		- conversion to Vector3f should return vertex position
	 */
	inline operator Vector3f() const
	{
		return m_p_ref->v_position;
	}

	/*
	 *	inline TRefVertex operator =(const Vector3f &r_v_position)
	 *		- assigning Vector3f should write to vertex position
	 */
	inline TRefVertex operator =(const Vector3f &r_v_position)
	{
		m_p_ref->v_position = r_v_position;
		return *this;
	}

	/*
	 *	inline bool operator ==(const TBaseVertex &r_t_vertex)
	 *		- we might need means to compare two vertices
	 *		- constant f_epsilon is used as maximal coordinate distance tolerance
	 */
	inline bool operator ==(const TBaseVertex &r_t_vertex)
	{
		return *m_p_ref == r_t_vertex;
	}

	/*
	 *	TVertex operator *(float f_scalar) const
	 *		- return vertex, scaled by f_scalar
	 *		- returns a new TBaseVertex because new vertex originated
	 *		  (in case it will need to be converted to TRefVertex, it will be added to the list)
	 */
	TBaseVertex operator *(float f_scalar) const
	{
		return *m_p_ref * f_scalar;
	}

	/*
	 *	TBaseVertex operator +(const TBaseVertex &r_vertex) const
	 *		- return vertex with sum of coordinates of both original vertices
	 *		- returns a new TBaseVertex because new vertex originated
	 *		  (in case it will need to be converted to TRefVertex, it will be added to the list)
	 */
	TBaseVertex operator +(const TBaseVertex &r_vertex) const
	{
		return *m_p_ref + r_vertex;
	}

	/*
	 *	TRefVertex operator *=(float f_scalar) const
	 *		- return vertex, scaled by f_scalar
	 *		- returns this because no new vertex originated
	 *		  (coordinates of some existing vertex were scaled only)
	 */
	TRefVertex operator *=(float f_scalar)
	{
		*m_p_ref *= f_scalar;
		return *this;
	}

	/*
	 *	TRefVertex operator +=(const TBaseVertex &r_vertex) const
	 *		- return vertex with sum of coordinates of both original vertices
	 *		- returns this because no new vertex originated
	 *		  (coordinates of some existing vertex were scaled only)
	 */
	TRefVertex operator +=(const TBaseVertex &r_vertex)
	{
		*m_p_ref += r_vertex;
		return *this;
	}
};

/*
 *	template <class TVertStruct>
 *	class CPolygon2
 *		- simple convex polygon template
 */
template <class TVertStruct>
class CPolygon2 {
protected:
	std::vector<TVertStruct> m_vertex_list;
	Plane3f m_t_normal;

public:
	/*
	 *	inline CPolygon2()
	 *		- default constructor
	 */
	inline CPolygon2()
	{}

	/*
	 *	CPolygon2(const CPolygon2<TVertStruct> &r_other)
	 *		- copy-constructor
	 */
	CPolygon2(const CPolygon2<TVertStruct> &r_other)
	{
		m_t_normal = r_other.m_t_normal;
		m_vertex_list.insert(m_vertex_list.begin(),
			r_other.m_vertex_list.begin(), r_other.m_vertex_list.end());
	}

	/*
	 *	bool operator =(const CPolygon2<TVertStruct> &r_other)
	 *		- copy operation. all current vertices are removed, new vertices and normal
	 *		  are copied from r_other polygon
	 *		- returns true on success or false in case there was not enough memory for vertices
	 */
	bool operator =(const CPolygon2<TVertStruct> &r_other)
	{
		m_t_normal = r_other.m_t_normal;
		// copy normal

		m_vertex_list.erase(m_vertex_list.begin(), m_vertex_list.end());
		m_vertex_list.reserve(r_other.m_vertex_list.size());
		if(m_vertex_list.capacity() < r_other.m_vertex_list.size())
			return false;
		// alloc vertex list

		m_vertex_list.insert(m_vertex_list.begin(),
			r_other.m_vertex_list.begin(), r_other.m_vertex_list.end());
		// copy vertex list

		return true;
	}

	/*
	 *	void Delete()
	 *		- delete all polygon vertices
	 */
	void Delete()
	{
		m_vertex_list.clear();
	}

	/*
	 *	inline bool Add_Vertex(const TVertStruct &r_t_vertex)
	 *		- add single vertex past the last vertex in the array
	 */
	inline bool Add_Vertex(const TVertStruct &r_t_vertex)
	{
		return Add_Vertex(m_vertex_list.size(), &r_t_vertex, 1);
	}

	/*
	 *	inline bool Add_Vertex(unsigned int n_insert_before, const TVertStruct &r_t_vertex)
	 *		- add single vertex before the n_insert_before-th vertex in the array
	 */
	inline bool Add_Vertex(unsigned int n_insert_before, const TVertStruct &r_t_vertex)
	{
		return Add_Vertex(n_insert_before, &r_t_vertex, 1);
	}

	/*
	 *	inline bool Add_Vertex(const TVertStruct *p_vertex, unsigned int n_count)
	 *		- add array of vertices past the last vertex in the array
	 */
	inline bool Add_Vertex(const TVertStruct *p_vertex, unsigned int n_count)
	{
		return Add_Vertex(m_vertex_list.size(), p_vertex, n_count);
	}

	/*
	 *	bool Add_Vertex(unsigned int n_insert_before, const TVertStruct *p_vertex, unsigned int n_count)
	 *		- add array of vertices before the n_insert_before-th vertex in the array
	 */
	bool Add_Vertex(unsigned int n_insert_before, const TVertStruct *p_vertex, unsigned int n_count)
	{
		//assert(n_count > 0);
		assert(/*n_insert_before >= 0 &&*/ n_insert_before <= m_vertex_list.size());
		m_vertex_list.reserve(m_vertex_list.size() + n_count);
		if(m_vertex_list.capacity() < m_vertex_list.size() + n_count)
			return false;
		m_vertex_list.insert(m_vertex_list.begin() + n_insert_before,
			p_vertex, p_vertex + n_count);
		return true;
	}

	/*
	 *	void Delete_Vertices(int n_index, int n_count)
	 *		- delete n_count vertices, beggining with vertex n_index
	 */
	void Delete_Vertices(int n_index, int n_count)
	{
		m_vertex_list.erase(&m_vertex_list[n_index], &m_vertex_list[n_index + n_count]);
	}

	/*
	 *	inline int n_Vertex_Num() const
	 *		- return number of vertices
	 */
	inline int n_Vertex_Num() const
	{
		return m_vertex_list.size();
	}

	/*
	 *	inline TVertStruct &r_t_Vertex(int n_index) const
	 *		- vertex access function
	 *		- doesn't check array bounds
	 */
	inline TVertStruct &r_t_Vertex(int n_index)
	{
		return m_vertex_list[n_index];
	}

	/*
	 *	inline TVertStruct &t_Vertex(int n_index) const
	 *		- vertex access function
	 *		- doesn't check array bounds
	 */
	inline const TVertStruct &t_Vertex(int n_index) const
	{
		return m_vertex_list[n_index];
	}

	/*
	 *	const Vector3f v_Center() const
	 *		- return position of center of polygon
	 *		- if polygon has no vertices, return the O vector
	 */
	const Vector3f v_Center() const
	{
		if(!m_vertex_list.size())
			return Vector3f(0, 0, 0);

		Vector3f v_center(0, 0, 0);
		for(unsigned int i = 0; i < m_vertex_list.size(); ++ i)
			v_center += m_vertex_list[i];

		return v_center * (1.0f / m_vertex_list.size());
	}

	/*
	 *	const typename TVertStruct::TIntermediate t_Center() const
	 *		- return vertex in center of polygon
	 *		- if polygon has no vertices, return the O vertex
	 */
	const typename TVertStruct::TIntermediate t_Center() const
	{
		TVertStruct::TIntermediate t_center(Vector3f(0, 0, 0));
		if(!m_vertex_list.size())
			return t_center;

		for(unsigned int i = 0; i < m_vertex_list.size(); ++ i)
			t_center += m_vertex_list[i];

		return t_center * (1.0f / m_vertex_list.size());
	}

	/*
	 *	bool Calc_Normal()
	 *		- calculate normal, return false if there wasn't trinity
	 *		  of vertices to construct plane from
	 *		- it has to be called explicitly, it is never called by CPolygon2 itself
	 */
	bool Calc_Normal()
	{
		if(m_vertex_list.size() < 3)
			return false;
		// impossible with so little vertices

		for(unsigned int i = 1; i < m_vertex_list.size(); ++ i) {
			Vector3f v_u = (Vector3f)m_vertex_list[i - 1] - (Vector3f)m_vertex_list[i];
			if(v_u.f_Length() < f_epsilon)
				continue;
			// get non-zero vector

			for(unsigned int j = 0; j < m_vertex_list.size(); ++ j) {
				if(j == i || j == i - 1)
					continue;
				Vector3f v_v = (Vector3f)m_vertex_list[j] - (Vector3f)m_vertex_list[i];
				if(v_v.f_Length() < f_epsilon)
					continue;
				// get another different non-zero vector

				Vector3f v_normal = v_u.v_Cross(v_v);
				if(v_normal.f_Length() < f_epsilon)
					continue;
				v_normal.Normalize();
				// colinear vectors would yield zero normal

				m_t_normal = Plane3f(m_vertex_list[0], v_normal);
				return true;
				// create plane
			}
		}

		return false;
	}

	/*
	 *	inline const Plane3f &t_Normal() const
	 *		- normal access function
	 */
	inline const Plane3f &t_Normal() const
	{
		return m_t_normal;
	}

	/*
	 *	template <class CFunctionObject>
	 *	inline CFunctionObject for_each_vertex(CFunctionObject t_object)
	 *		- for-each cycle over polygon's vertices
	 */
	template <class CFunctionObject>
	inline CFunctionObject for_each_vertex(CFunctionObject t_object)
	{
		return std::for_each(m_vertex_list.begin(), m_vertex_list.end(), t_object);
	}

	/*
	 *	template <class CFunctionObject>
	 *	inline CFunctionObject for_each_vertex(CFunctionObject t_object) const
	 *		- for-each cycle over polygon's const vertices
	 */
	template <class CFunctionObject>
	inline CFunctionObject for_each_vertex(CFunctionObject t_object) const
	{
		return std::for_each(m_vertex_list.begin(), m_vertex_list.end(), t_object);
	}

	/*
	 *	void Swap_VertexOrder()
	 *		- swap vertex order, flip normal
	 */
	void Swap_VertexOrder()
	{
		m_t_normal.v_normal *= -1;
		m_t_normal.f_dist *= -1;

		for(unsigned int i = 0; i < m_vertex_list.size() / 2; ++ i) {
			TVertStruct t_tmp(m_vertex_list[i]);
			m_vertex_list[i] = m_vertex_list[m_vertex_list.size() - 1 - i];
			m_vertex_list[m_vertex_list.size() - 1 - i] = t_tmp;
		}
		// swap vertices
	}

protected:
	/*
	 *	class CFindConcave
	 *		- small function object, finds vertex, forming concave edge
	 */
	class CFindConcave {
	protected:
		Vector3f m_v_prev;
		Vector3f m_v_prev2;
		Vector3f m_v_normal;
		Vector3f m_v_u;

	public:
		inline CFindConcave(Vector3f v_prev, Vector3f v_prev2, Vector3f v_normal)
			:m_v_prev(v_prev), m_v_prev2(v_prev2), m_v_normal(v_normal)
		{
			m_v_u = m_v_prev - m_v_prev2;
			m_v_u.Normalize();
		}

		inline bool operator ()(const Vector3f &r_v_vertex_pos)
		{
			Vector3f v_v = m_v_prev2 - r_v_vertex_pos;
			v_v.Normalize();

			if(m_v_u.v_Cross(m_v_normal).f_Dot(v_v - m_v_u) < -f_epsilon)
				return true;
			
			m_v_prev = m_v_prev2;
			m_v_prev2 = r_v_vertex_pos;
			m_v_u = v_v;
			// shift vectors

			return false;
		}
	};

public:
	/*
	 *	inline bool b_IsConvex() const
	 *		- returns true in case polygon is convex, otherwise returns false
	 */
	inline bool b_IsConvex() const
	{
		if(m_vertex_list.size() < 3)
			return true;
		// polygon with so little vertices can't be concave

		return std::find_if(m_vertex_list.begin(),
			m_vertex_list.end(), CFindConcave(m_vertex_list[m_vertex_list.size() - 2],
			m_vertex_list[m_vertex_list.size() - 1],
			m_t_normal.v_normal)) == m_vertex_list.end();
		// in case concave edge is not found, it's convex polygon

		/*Vector3f v_u = (Vector3f)m_vertex_list[m_vertex_list.size() - 1] -
			(Vector3f)m_vertex_list[0];
		v_u.Normalize();
		for(unsigned int i = 0; i < m_vertex_list.size(); ++ i) {
			Vector3f v_v = (Vector3f)m_vertex_list[i] -
				(Vector3f)m_vertex_list[(i + 1) % m_vertex_list.size()];
			v_v.Normalize();
			// get two edge vectors ...

			if(v_u.v_Cross(m_t_normal.v_normal).f_Dot(v_v - v_u) < -f_epsilon)
				return false;
			// see if difference vectors goes to the same side as cross
			// product of first edge vector with normal

			v_u = v_v;
		}
		return true;*/
		// old function using cycle
	}

protected:
	/*
	 *	class CFindMaxPlaneDeviation
	 *		- simple function object for finding maximal
	 *	      absolute vertex deviation from normal plane
	 */
	class CFindMaxPlaneDeviation {
	protected:
		float m_f_max_deviation;
		const Plane3f &m_r_plane;

	public:
		inline CFindMaxPlaneDeviation(const Plane3f &r_plane)
			:m_r_plane(r_plane)
		{}

		inline void operator ()(const Vector3f &r_v_vertex_pos)
		{
			float f_deviation = (float)fabs(m_r_plane.f_Vector_Dist(r_v_vertex_pos));
			if(m_f_max_deviation < f_deviation)
				m_f_max_deviation = f_deviation;
		}

		inline operator float() const
		{
			return m_f_max_deviation;
		}
	};

public:
	/*
	 *	inline float f_MaxPlaneDeviation() const
	 *		- returns maximal absolute distance of polygon points from polygon plane
	 */
	inline float f_MaxPlaneDeviation() const
	{
		return std::for_each(m_vertex_list.begin(),
			m_vertex_list.end(), CFindMaxPlaneDeviation(m_t_normal));
	}

protected:
	/*
	 *	class CFindLongEdges
	 *		- small function object used to find third edge longer than epsilon
	 *		  (epsilon is explicitly specified in constructor)
	 */
	class CFindLongEdges {
	protected:
		Vector3f m_v_prev;
		float m_f_epsilon;
		int m_n_edge_num;

	public:
		inline CFindLongEdges(Vector3f v_prev, float f_epsilon_ex)
			:m_v_prev(v_prev), m_f_epsilon(f_epsilon_ex), m_n_edge_num(0)
		{}

		inline bool operator ()(Vector3f v_vertex)
		{
			if((m_v_prev - v_vertex).f_Length() >= m_f_epsilon) {
				if(m_n_edge_num == 2)
					return true;
				// in case there are three edges, triangle is not tiny

				++ m_n_edge_num;
			}
			// is the edge longer than epsilon?

			m_v_prev = v_vertex;

			return false;
		}
	};

public:
	/*
	 *	inline bool b_IsTiny() const
	 *		- returns true in case polygon doesn't contain at
	 *		  least three edges longer than epsilon, otherwise false
	 */
	inline bool b_IsTiny() const
	{
		if(m_vertex_list.size() < 3)
			return true;
		return std::find_if(m_vertex_list.begin(),
			m_vertex_list.end(), CFindLongEdges(m_vertex_list[m_vertex_list.size() - 1],
			f_epsilon)) == m_vertex_list.end();
	}

protected:
	/*
	 *	class CEvalPlanePos
	 *		- small function object for evaluating polygon position against plane
	 */
	class CEvalPlanePos {
	protected:
		unsigned int m_n_front_num;
		unsigned int m_n_back_num;
		const Plane3f &m_r_t_plane;

	public:
		inline CEvalPlanePos(const Plane3f &r_t_plane)
			:m_r_t_plane(r_t_plane), m_n_front_num(0), m_n_back_num(0)
		{}

		inline void operator ()(const Vector3f &r_v_vertex) const
		{
			EPlanePos n_pos;
			if((n_pos = m_r_t_plane.n_Vector_Pos(r_v_vertex)) == plane_Front)
				++ m_n_front_num;
			else if(n_pos == plane_Back)
				++ m_n_back_num;
			// determine vertex side and increment counters accordingly

			return false;
		}

		inline operator EPlanePos() const
		{
			if(m_n_front_num) {
				if(!m_n_back_num)
					return plane_Front; // front propably occurres more frequently
				else
					return plane_Split;
			} else {
				if(m_n_front_num)
					return plane_Split; // split propably occurres more frequently
				else
					return plane_Onplane;
			}
		}
	};

public:
	/*
	 *	inline EPlanePos n_Plane_Pos(const Plane3f &r_t_plane) const
	 *		- returns relative position of polygon against plane r_t_plane
	 *		  (one of plane_Front, plane_Back, plane_Split or plane_Onplane)
	 *		- note polygons with no vertices will be classified as onplane
	 */
	inline EPlanePos n_Plane_Pos(const Plane3f &r_t_plane) const
	{
		return std::for_each(m_vertex_list.begin(),
			m_vertex_list.end(), CEvalPlanePos(r_t_plane));
	}

	/*
	 *	bool b_PointInside_SumAngles(Vector3f v_point) const
	 *		- returns true in case point v_point is inside the polygon
	 *		- uses method of summing up point-to-edge vector angles
	 */
	bool b_PointInside_SumAngles(Vector3f v_point) const
	{
		float f_angle_sum = 0;
		
		if(!m_vertex_list.size())
			return false;

		Vector3f v_prev = (Vector3f)m_vertex_list[m_vertex_list.size() - 1] - v_point;
		v_prev.Normalize();
		for(unsigned int i = 0; i < m_vertex_list.size(); ++ i) {
			Vector3f v_cur = (Vector3f)m_vertex_list[i] - v_point;
			v_cur.Normalize();
			// take two vectors ...

			f_angle_sum += (float)acos(v_prev.f_Dot(v_cur));
			// sum up the angles they hold

			v_prev = v_cur;
		}
		// sum up the angles

		return fabs(f_angle_sum - 2 * f_pi) < f_epsilon;
		// in case difference from 2pi is less than epsilon, the point is inside
	}

	/*
	 *	bool b_PointInside_EdgePlanes(Vector3f v_point) const
	 *		- returns true in case point v_point is inside the polygon
	 *		- uses method of edge planes 'fence'
	 */
	bool b_PointInside_EdgePlanes(Vector3f v_point) const
	{
		for(unsigned int i = 0; i < m_vertex_list.size(); ++ i) {
			Vector3f v_u = (Vector3f)m_vertex_list[i] -
				(Vector3f)m_vertex_list[(i + 1) % m_vertex_list.size()];
			// get an edge vector ...

			Vector3f v_v = v_point - (Vector3f)m_vertex_list[i];
			// get vector to point

			if(v_u.v_Cross(m_t_normal.v_normal).f_Dot(v_v) > 0)
				return false;
			// see if the point lies behind the 'fence' plane
			// (in case it does, it's outside)
		}

		return true;
	}

	/*
	 *	bool b_PointInside_Barycentric(Vector3f v_point) const
	 *		- returns true in case point v_point is inside the polygon
	 *		- uses barycentric coordinates and polygon to triangles decomposition
	 */
	bool b_PointInside_Barycentric(Vector3f v_point) const
	{
		for(int i = 1; i < (int)m_vertex_list.size() - 1; ++ i) {
			const TVertStruct *p_vert0 = &m_vertex_list[0];
			const TVertStruct *p_vert1 = &m_vertex_list[i];
			const TVertStruct *p_vert2 = &m_vertex_list[(i + 1) % m_vertex_list.size()];

			Vector3f v_edge1 = (Vector3f)*p_vert1 - (Vector3f)*p_vert0,
					 v_edge2 = (Vector3f)*p_vert2 - (Vector3f)*p_vert0;
			float f_inv_length2_edge1 = 1.0f / v_edge1.f_Length2();
			Vector3f v_perp = v_edge2 - v_edge1 * (v_edge1.f_Dot(v_edge2) * f_inv_length2_edge1);
			Vector3f v_t = v_point - (Vector3f)*p_vert0;
			float f_u = v_t.f_Dot(v_perp) / v_perp.f_Dot(v_edge2);
			// calculate barycentric coordinates for v_point

			if(f_u < 0)
				continue;
			// first condition failed

			float f_v = (v_t - v_edge2 * f_u).f_Dot(v_edge1) * f_inv_length2_edge1;
			if(f_v >= 0 && f_u + f_v <= 1)
				return true;
			// in case the other conditions are satisfied as well, point lies inside
		}

		return false;
	}

protected:
	/*
	 *	class CTestVertices
	 *		- little function object for finding vertex, lying inside triangle specified
	 *		  by one vertex two of it's edges sharing the vertex, vector perpendicular
	 *		  to edge1 and inverse squared length of edge1 (barrycentric coords method)
	 */
	class CTestVertices {
	protected:
		Vector3f m_v_vert0, m_v_edge1, m_v_edge2, m_v_perp;
		float m_f_inv_length2_edge1;

	public:
		inline CTestVertices(Vector3f v_vert0, Vector3f v_edge1,
			Vector3f v_edge2, Vector3f v_perp, float f_inv_length2_edge1)
			:m_v_vert0(v_vert0), m_v_edge1(v_edge1), m_v_edge2(v_edge2), m_v_perp(v_perp),
			m_f_inv_length2_edge1(f_inv_length2_edge1)
		{}

		inline bool operator ()(const Vector3f &r_v_pos) const
		{
			Vector3f v_t = r_v_pos - m_v_vert0;
			float f_u = v_t.f_Dot(m_v_perp) / m_v_perp.f_Dot(m_v_edge2);
			// calculate barycentric coordinates for v_point

			if(f_u < 0)
				return false;
			// first condition failed

			float f_v = (v_t - m_v_edge2 * f_u).f_Dot(m_v_edge1) * m_f_inv_length2_edge1;
			return f_v >= 0 && f_u + f_v <= 1;
			// in case the other conditions are satisfied as well, point lies inside
		}
	};

public:
	/*
	 *	template <class CPolygon2/*<class TVertStruct2>* / >
	 *	bool b_Overlap_Barycentric(const CPolygon2/*<TVertStruct2>* / &r_poly2) const
	 *		- returns true in case r_poly2 overlaps this polygon, otherwise false
	 *		- uses barycentric coordinates test for determining wheter vertices
	 *		  from one polygon lies in the other one and vice verse
	 *		- note second polygon can be of another vertex type, this comes in handy
	 *		  when dealing with portals
	 */
	template <class CPolygon2_TVertStruct2_ >
	bool b_Overlap_Barycentric(const CPolygon2_TVertStruct2_ &r_poly2) const
	{
		for(int i = 1; i < (int)m_vertex_list.size() - 1; ++ i) {
			const TVertStruct *p_vert0 = &m_vertex_list[0];
			const TVertStruct *p_vert1 = &m_vertex_list[i];
			const TVertStruct *p_vert2 = &m_vertex_list[(i + 1) % m_vertex_list.size()];

			Vector3f v_edge1 = (Vector3f)*p_vert1 - (Vector3f)*p_vert0,
					 v_edge2 = (Vector3f)*p_vert2 - (Vector3f)*p_vert0;
			float f_inv_length2_edge1 = 1.0f / v_edge1.f_Length2();
			Vector3f v_perp = v_edge2 - v_edge1 * (v_edge1.f_Dot(v_edge2) * f_inv_length2_edge1);

			if(std::find_if(r_poly2.m_vertex_list.begin(), r_poly2.m_vertex_list.end(),
			   CTestVertices(*p_vert0, v_edge1, v_edge2, v_perp, f_inv_length2_edge1)) <
			   r_poly2.m_vertex_list.end())
				return true;
			// see if any of those vertices lies in this part of polygon
		}
		// see if any of r_poly2's vertices lies in this polygon

		for(int i = 1; i < (int)r_poly2.m_vertex_list.size() - 1; ++ i) {
			const TVertStruct *p_vert0 = &r_poly2.m_vertex_list[0];
			const TVertStruct *p_vert1 = &r_poly2.m_vertex_list[i];
			const TVertStruct *p_vert2 = &r_poly2.m_vertex_list[(i + 1) %
				r_poly2.m_vertex_list.size()];

			Vector3f v_edge1 = (Vector3f)*p_vert1 - (Vector3f)*p_vert0,
					 v_edge2 = (Vector3f)*p_vert2 - (Vector3f)*p_vert0;
			float f_inv_length2_edge1 = 1.0f / v_edge1.f_Length2();
			Vector3f v_perp = v_edge2 - v_edge1 * (v_edge1.f_Dot(v_edge2) *
				f_inv_length2_edge1);

			if(std::find_if(m_vertex_list.begin(), m_vertex_list.end(),
			   CTestVertices(*p_vert0, v_edge1, v_edge2, v_perp, f_inv_length2_edge1)) <
			   m_vertex_list.end())
				return true;
			// see if any of those vertices lies in this part of polygon
		}
		// see if any of this polygon's vertices lies in r_poly2

		return false;
	}

	/*
	 *	Vector3f v_NearestPoint_Barycentric(Vector3f &r_v_vector) const
	 *		- returns nearest point to r_v_vector contained by polygon
	 */
	Vector3f v_NearestPoint_Barycentric(Vector3f &r_v_vector) const
	{
		Vector3f v_projection = r_v_vector - m_t_normal.v_normal *
			(m_t_normal.v_normal.f_Dot(r_v_vector) + m_t_normal.f_dist);
		if(b_PointInside_Barycentric(v_projection))
			return v_projection;
		// calculate projection, then in case it's inside, return it

		/*
		if(b_PointInside_EdgePlanes(r_v_vector)) {
			return r_v_vector - m_t_normal.v_normal *
				(m_t_normal.v_normal.f_Dot(r_v_vector) + m_t_normal.f_dist);
		}
		// in case it's inside, calculate projection and return it
		*/ // possibly faster version using edge planes 'fence'

		Vector3f v_closest;
		float f_min_dist2 = 1e10f;
		for(unsigned int i = 0; i < m_vertex_list.size(); ++ i) {
			Vector3f v_u = (Vector3f)m_vertex_list[(i + 1) % m_vertex_list.size()] -
				(Vector3f)m_vertex_list[i];
			// get an edge vector ...

			Vector3f v_v = r_v_vector - (Vector3f)m_vertex_list[i];
			// get vector from edge origin to point

			float f_t = v_u.f_Dot(v_v) / v_u.f_Length2();
			//float f_t = v_u.f_Dot(v_v) / v_u.f_Dot(v_u); // equivalent to the previous one
			// get projection distance along the edge

			Vector3f v_projection;
			if(f_t < 0)
				v_projection = m_vertex_list[i];
			else if(f_t > 1)
				v_projection = m_vertex_list[(i + 1) % m_vertex_list.size()];
			else
				v_projection = (Vector3f)m_vertex_list[i] + v_u * f_t;
			// get projection position

			float f_dist2 = (r_v_vector - v_projection).f_Length2();
			if(f_dist2 < f_min_dist2 || !i) {
				f_min_dist2 = f_dist2;
				v_closest = v_projection;
			}
			// find the closest projection
		}

		return v_closest;
	}

	/*
	 *	bool Cut(const Plane3f &r_t_plane, EPlanePos n_desired_half = plane_Front)
	 *		- cuts polygon by plane r_t_plane so only the half lying on n_desired_half side
	 *		  will remain. note plane_Front and plane_Back are the only allowed values
	 *		- returns true in case of success or false in case there was not enough memory
	 */
	bool Cut(const Plane3f &r_t_plane, EPlanePos n_desired_half = plane_Front)
	{
		//assert(n_desired_half == plane_Front || n_desired_half == plane_Back);

		if(!m_vertex_list.size())
			return true;
		// nothing to do here

		EPlanePos n_other_half = (n_desired_half == plane_Front)? plane_Back : plane_Front;
		// the other ('wrong one') position

		TVertStruct t_prev_vertex = m_vertex_list[m_vertex_list.size() - 1];
		bool b_prev_good_side = r_t_plane.n_Vector_Pos(t_prev_vertex) != n_other_half;
		for(unsigned int i = 0; i < m_vertex_list.size(); ++ i) {
			TVertStruct t_vertex = m_vertex_list[i];
			bool b_good_side = r_t_plane.n_Vector_Pos(t_vertex) != n_other_half;
			// determine plane sides previous and current vertex lies on

			if(b_good_side == b_prev_good_side) {
				if(b_good_side) {
					// both sides in front, keep vertex in the list 'as is'
				} else {
					// both sides in back, remove current vertex from the list

					m_vertex_list.erase(m_vertex_list.begin() + i,
						m_vertex_list.begin() + i + 1);
					-- i;
				}
			} else {
				float f_t = r_t_plane.f_Vector_Dist(t_vertex);
				float f_subdivide_tresh = 1.0f / (f_t - r_t_plane.f_Vector_Dist(t_prev_vertex));
				f_t *= f_subdivide_tresh;
				// calculate distance of subdivision along the edge

				if(f_subdivide_tresh < 0)
					f_subdivide_tresh = -f_subdivide_tresh;
				f_subdivide_tresh *= f_epsilon;
				// calculate treshold distance

				if(f_t > f_subdivide_tresh &&
				   f_t < 1 - f_subdivide_tresh) {
					// new vertex is far enough from existing vertices ...
					// (prevents from creating tiny edges)

					TVertStruct t_split_vertex(t_vertex);
					t_split_vertex = t_vertex * (1 - f_t) + t_prev_vertex * f_t;
					// calculate a new vertex lying on edge in place where it's cut by the plane

					if(b_good_side) {
						// prev side was on the wrong position of plane, we have to insert
						// split vertex before current vertex

						if(!Add_Vertex(i, t_split_vertex))
							return false;
						++ i;
					} else {
						// this side is on the wrong position, replace the current vertex by
						// split vertex

						m_vertex_list[i] = t_split_vertex;
					}
				} else if(!b_good_side) {
					m_vertex_list.erase(m_vertex_list.begin() + i,
						m_vertex_list.begin() + i + 1);
					-- i;
				}
			}

			b_prev_good_side = b_good_side;
			t_prev_vertex = t_vertex;
		}
		// cut polygon

		return true;
	}

	/*
	 *	bool Split(const Plane3f &r_t_plane, CPolygon2<TVertStruct> &r_other,
	 *		EPlanePos n_desired_half = plane_Front)
	 *		- splits polygon by plane r_t_plane so only the half lying on n_desired_half side
	 *		  will remain. the other half is put inside r_other polygon. note plane_Front and
	 *		  plane_Back are the only allowed values
	 *		- returns true in case of success or false in case there was not enough memory
	 *		- note in case this becomes tiny polygon after operation, it's propable that r_other
	 *		  contains double vertex which is looked up and removed (TVertStruct should
	 *		  therefore either not implement operator == at all or it should be taken as vertex
	 *		  identity (no epsilons in it))
	 */
	bool Split(const Plane3f &r_t_plane, CPolygon2<TVertStruct> &r_other,
		EPlanePos n_desired_half = plane_Front)
	{
		//assert(n_desired_half == plane_Front || n_desired_half == plane_Back);

		r_other.m_t_normal = m_t_normal;
		// copy normal

		r_other.m_vertex_list.erase(r_other.m_vertex_list.begin(), r_other.m_vertex_list.end());
		// erase all the vertices

		if(!m_vertex_list.size())
			return true;
		// nothing to do here

		EPlanePos n_other_half = (n_desired_half == plane_Front)? plane_Back : plane_Front;
		// the other ('wrong one') position

		TVertStruct t_prev_vertex = m_vertex_list[m_vertex_list.size() - 1];
		bool b_prev_good_side = r_t_plane.n_Vector_Pos(t_prev_vertex) != n_other_half;
		//bool b_prev_bad_side = r_t_plane.n_Vector_Pos(t_prev_vertex) != n_desired_half;
		for(unsigned int i = 0; i < m_vertex_list.size(); ++ i) {
			TVertStruct t_vertex = m_vertex_list[i];
			bool b_good_side = r_t_plane.n_Vector_Pos(t_vertex) != n_other_half;
			//bool b_bad_side = r_t_plane.n_Vector_Pos(t_vertex) != n_desired_half;
			// determine plane sides previous and current vertex lies on

			if(b_good_side == b_prev_good_side) {
				if(b_good_side) {
					// both sides in front, keep vertex in the list 'as is'
				} else {
					// both sides in back, remove current vertex from the list

					if(!r_other.Add_Vertex(t_vertex))
						return false;
					m_vertex_list.erase(m_vertex_list.begin() + i,
						m_vertex_list.begin() + i + 1);
					-- i;
				}
			} else {
				float f_t = r_t_plane.f_Vector_Dist(t_vertex);
				float f_subdivide_tresh = 1.0f / (f_t - r_t_plane.f_Vector_Dist(t_prev_vertex));
				f_t *= f_subdivide_tresh;
				// calculate distance of subdivision along the edge

				if(f_subdivide_tresh < 0)
					f_subdivide_tresh = -f_subdivide_tresh;
				f_subdivide_tresh *= f_epsilon;
				// calculate treshold distance

				if(f_t > f_subdivide_tresh &&
				   f_t < 1 - f_subdivide_tresh) {
					// new vertex is far enough from existing vertices ...
					// (prevents from creating tiny edges)

					TVertStruct t_split_vertex(t_vertex);
					t_split_vertex = t_vertex * (1 - f_t) + t_prev_vertex * f_t;
					// calculate a new vertex lying on edge in place where it's cut by the plane

					if(b_good_side) {
						// prev side was on the wrong position of plane, we have to insert
						// split vertex before current vertex

						if(!Add_Vertex(i, t_split_vertex) ||
						   !r_other.Add_Vertex(t_split_vertex))
							return false;
						++ i;
					} else {
						// this side is on the wrong position, replace the current vertex by
						// split vertex

						if(!r_other.Add_Vertex(t_split_vertex) ||
						   !r_other.Add_Vertex(t_vertex))
							return false;
						m_vertex_list[i] = t_split_vertex;
					}
				} else {
					if(b_good_side) {
						if(!r_other.Add_Vertex(t_vertex))
							return false;
					} else {
						if(!r_other.Add_Vertex(t_prev_vertex) ||
						   !r_other.Add_Vertex(t_vertex))
							return false;
						m_vertex_list.erase(m_vertex_list.begin() + i,
							m_vertex_list.begin() + i + 1);
						-- i;
					}
				}
			}

			//b_prev_bad_side = b_bad_side;
			b_prev_good_side = b_good_side;
			t_prev_vertex = t_vertex;
		}
		// cut polygon

		if(m_vertex_list.size() == 1) {
			TVertStruct t_prev_vertex = r_other.m_vertex_list[r_other.m_vertex_list.size() - 1];
			for(int i = 0; i < r_other.m_vertex_list.size(); ++ i) {
				if(t_prev_vertex == r_other.m_vertex_list[i]) {
					r_other.m_vertex_list.erase(r_other.m_vertex_list.begin() + i,
						r_other.m_vertex_list.begin() + i + 1);
					return true;
				}
				t_prev_vertex = r_other.m_vertex_list[i];
			}
		}
		// in case there is just a single vertex left in this polygon,
		// it was added as double vertex to r_other

		return true;
	}
};

#endif //__POLYGON_INCLUDED
